import logging
from itsdangerous import BadSignature
from tornado import ioloop, websocket
import datetime
import uuid


logger = logging.getLogger('tornadopush')


class ChannelMixin(object):
    @property
    def channel(self):
        return getattr(self, '_channel', None)

    @channel.setter
    def channel(self, channel):
        if not channel:
            channel = self.application.settings['default_channel']
        allowed_channels = self.application.settings['allowed_channels']
        if allowed_channels and channel not in allowed_channels:
            self.send_error(404)
        self._channel = channel

    @property
    def channel_key(self):
        return self.application.settings['channel_format'] % self.channel


class AuthentifyMixin(object):
    @property
    def auth_enabled(self):
        return self.application.settings['auth_enabled']

    def authentify(self, token=None, callback=None):
        self.user_id = None
        self.user_channels = None
        if not callback:
            callback = self.on_authorized
        if token:
            self._validate_token(token)
        if self.user_id is None and self.auth_enabled:
            logger.debug('Token %s rejected' % token)
            self.on_unauthorized()
        else:
            if self.user_id:
                logger.debug('User %s connected on #%s' % (self.user_id, self.channel))
            else:
                logger.debug('New anonymous connected on #%s' % self.channel)
                self.user_id = str(uuid.uuid4())
            callback()

    def _validate_token(self, token):
        max_age = self.application.settings['token_max_age']
        try:
            user_id, user_channels = self.application.token_serializer.loads(token, max_age=max_age)
            if not user_channels or not self.channel or self.channel in user_channels:
                self.user_id = user_id
                self.user_channels = user_channels
        except BadSignature as e:
            pass

    def on_unauthorized(self):
        self.send_error(403)

    def on_authorized(self):
        self.auth_done = True


class WebSocketMixin(object):
    heartbeat_interval = 60
    heartbeat_timeout = 30

    def send_error(self, code, **kwargs):
        self.close()

    def check_origin(self, origin):
        return True

    def start_heartbeat(self, interval=None, timeout=None):
        self.heartbeat = True
        if interval is not None:
            self.heartbeat_interval = interval
        if timeout is not None:
            self.heartbeat_timeout = timeout

        def close_on_timeout():
            logger.debug('Ping timeout')
            self.heartbeat_timeout_cb = None
            self.heartbeat = False
            self.close()
            self.on_close()

        self.heartbeat_timeout_cb = ioloop.IOLoop.instance()\
            .call_later(self.heartbeat_timeout, close_on_timeout)
        logger.debug('Pinging...')
        try:
            self.ping('ping')
        except websocket.WebSocketClosedError:
            logger.debug('Client already disconnected')
            ioloop.IOLoop.instance().remove_timeout(self.heartbeat_timeout_cb)
            self.heartbeat_timeout_cb = None
            close_on_timeout()

    def on_pong(self, data):
        logger.debug('Pong')
        if self.heartbeat_timeout_cb:
            ioloop.IOLoop.instance().remove_timeout(self.heartbeat_timeout_cb)
            self.heartbeat_timeout_cb = None
        if getattr(self, 'heartbeat', False):
            ioloop.IOLoop.instance().call_later(
                self.heartbeat_interval, self.start_heartbeat)



import json
from itsdangerous import URLSafeTimedSerializer
from redis import StrictRedis


class Channel(object):
    def __init__(self, client, channel):
        self.client = client
        self.channel = channel

    def emit(self, event, data=None, target=None, serialize=True):
        if serialize:
            data = json.dumps(data)
        event_data = '%s;%s;%s' % (event, target or "", data)
        self.client.publish(self.channel, event_data)


class EventEmitter(object):
    def __init__(self, client, secret_key='tornadopush'):
        self.client = client
        self.token_serializer = URLSafeTimedSerializer(secret_key)

    def create_token(self, user_id, allowed_channels=None):
        return self.token_serializer.dumps([user_id, allowed_channels])

    def channel(self, name):
        return Channel(self.client, name)

    def emit(self, channel, event, data=None, target=None, serialize=True):
        self.channel(channel).emit(event, data, target, serialize)


_emitter = None

def connect(url=None, secret_key='tornadopush'):
    global _emitter
    if isinstance(url, StrictRedis):
        client = url
    elif url:
        client = StrictRedis.from_url(url)
    else:
        client = StrictRedis()
    _emitter = EventEmitter(client, secret_key)


def emit(channel, event, data=None, target=None, serialize=True):
    _emitter.emit(channel, event, data, target, serialize)


import tornadoredis
import toredis
from .base import logger


class BaseClient(object):
    def __init__(self):
        self.disconnect_callback = None
        self.subscribed = False
        self.subscribed_to = []

    def subscribe(self, channels, listener, callback=None):
        raise NotImplementedError

    def unsubscribe(self, callback=None):
        pass

    def close(self, callback=None):
        self.unsubscribe(callback)

    def set_disconnect_callback(self, callback):
        self.disconnect_callback = callback


class TornadoRedisQueueClient(BaseClient):
    def __init__(self, app):
        super(TornadoRedisQueueClient, self).__init__()
        self.listener = None
        self.client = tornadoredis.Client(app.settings['redis_host'], app.settings['redis_port'],
            app.settings['redis_password'], app.settings['redis_db'])
        self.client.connect()

    def subscribe(self, channels, listener, callback=None):
        self.listener = listener
        def on_subscribe(*args):
            self.client.listen(self.on_message)
            self.subscribed = True
            self.subscribed_to = channels
            if callback:
                callback()
        self.client.subscribe(channels, on_subscribe)

    def on_message(self, msg):
        if msg.kind == 'disconnect':
            logger.debug('Listener has received disconnect')
            if self.disconnect_callback:
                self.disconnect_callback()
            return
        if msg.kind in ('subscribe', 'unsubscribe'):
            return
        if self.listener:
            self.listener(str(msg.body))

    def unsubscribe(self, callback=None):
        if not self.subscribed:
            if callback:
                callback()
            return
        def cb():
            self.subscribed = False
            if callback:
                callback()
        self.client.unsubscribe(self.subscribed_to, callback=cb)

    def close(self, callback=None):
        def on_unsubscribe(*args):
            self.client.disconnect()
            logger.debug('Client disconnected')
            if callback:
                callback()
        if self.subscribed:
            self.unsubscribe(on_unsubscribe)
        else:
            on_unsubscribe()


class TornadoRedisQueue(object):
    def __init__(self, app):
        self.app = app

    def create_client(self):
        return TornadoRedisQueueClient(self.app)

    def publish(self, channel, message):
        pass


class ToRedisQueueClient(BaseClient):
    def __init__(self, app):
        super(ToRedisQueueClient, self).__init__()
        self.listener = None
        self.client = toredis.Client()
        self.client.connect(host=app.settings['redis_host'], port=app.settings['redis_port'])
        self.closing = False

    def subscribe(self, channels, listener, callback=None):
        self.listener = listener
        self.client.subscribe(channels, callback=self.on_message)
        self.subscribed = True
        self.subscribed_to = channels
        if callback:
            callback()

    def on_message(self, msg):
        if not msg:
            if self.closing:
                return
            logger.debug('Listener has received disconnect')
            if self.disconnect_callback:
                self.disconnect_callback()
            return
        msg_type, msg_channel, msg_value = msg
        if msg_type == 'message' and self.listener:
            self.listener(str(msg_value))

    def unsubscribe(self, callback=None):
        if not self.subscribed:
            if callback:
                callback()
            return
        def cb():
            self.subscribed = False
            if callback:
                callback()
        self.client.unsubscribe(self.subscribed_to, callback=cb)

    def close(self, callback=None):
        self.closing = True
        self.client._stream.close()
        logger.debug('Client disconnected')
        if callback:
            callback()


class ToRedisQueue(object):
    def __init__(self, app):
        self.app = app

    def create_client(self):
        return ToRedisQueueClient(self.app)

    def publish(self, channel, message):
        pass


class LocalQueueClient(BaseClient):
    def __init__(self, queue):
        super(LocalQueueClient, self).__init__()
        self.queue = queue
        self.listener = None

    def subscribe(self, channels, listener, callback=None):
        self.listener = listener
        for channel in channels:
            self.queue.channels.setdefault(channel, []).append(listener)
        self.subscribed = True
        self.subscribed_to = channels
        if callback:
            callback()

    def unsubscribe(self, callback=None):
        for channel in self.subscribed_to:
            self.queue.channels.get(channel, []).remove(self.listener)
        self.subscribed = False
        if callback:
            callback()


class LocalQueue(object):
    def __init__(self, app):
        self.app = app
        self.channels = {}

    def create_client(self):
        return LocalQueueClient(self)

    def publish(self, channel, message):
        for listener in self.channels.get(channel, []):
            listener(message)


queue_providers = {"tornado_redis": TornadoRedisQueue,
                   "toredis": ToRedisQueue,
                   "local": LocalQueue}


class QueueMixin(object):
    queue_client = None

    def subscribe(self, *channels):
        self.queue_client = self.application.queue_provider.create_client()
        self.queue_client.set_disconnect_callback(self.on_queue_disconnect)
        self.queue_client.subscribe(channels, self.on_queue_message)

    def unsubscribe(self, callback=None):
        if self.queue_client:
            self.queue_client.unsubscribe(callback)

    def publish_queue_message(self, channel, msg):
        self.application.queue_provider.publish(channel, msg)

    def on_queue_disconnect(self):
        pass

    def close_queue_client(self, callback=None):
        if self.queue_client:
            self.queue_client.close(callback)



from setuptools import setup, find_packages


setup(
    name='tornadopush',
    version='0.6.2',
    url='http://github.com/frascoweb/tornadopush',
    license='MIT',
    author='Maxime Bouroumeau-Fuseau',
    author_email='maxime.bouroumeau@gmail.com',
    description='Push and presence server built with Tornado and Redis',
    packages=find_packages(),
    package_data={
        'tornadopush': ['static/*.js'],
    },
    zip_safe=False,
    platforms='any',
    install_requires=[
        'tornado>=4.0.2',
        'redis>=2.10.3',
        'itsdangerous>=0.24',
        'PyYAML>=3.11',
        'jsmin>=2.0.11',
        'tornado-redis>=2.4.18',
        'toredis==0.1.2'
    ],
    entry_points='''
        [console_scripts]
        tornadopush=tornadopush.cli:main
    '''
)
